// Aner Ben-Artzi - Efficient Coherence-Based Visibility Sampling (CBVS) of Environment Maps
// Columbia University 2004
//
// Area.cpp: implementation of the Area class.
//
//////////////////////////////////////////////////////////////////////

#include "Area.h"
#include "Container.h"
#include "VVTrace.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Area::Area(int thesize)
{
	size = thesize;
	area = new Point3d[size*size];
	activex = size/2;
	activey = size/2;
	textx = 0;
	texty = size-1;
	numhighlighted = 0;
	activeLight = 0;
	type = TRACED;
	trace = NULL;
	normalize = false;
	closeup = false;
	corner = 0;
	highlightyellow = false;
	parentArea = NULL;
	lockFocus = false;
}

Area::~Area()
{
	delete[] area;
}

void Area::unFocus()
{
	int nh;
	for (nh = 0; nh < numhighlighted; nh++) {
		if (closeup) {
			(*highlighted[nh])[0] -= 10;
			(*highlighted[nh])[1] += 10;
			(*highlighted[nh])[2] += 10;
		} else if (highlightyellow) {
			(*highlighted[nh])[0] -= 10;
			(*highlighted[nh])[1] -= 10;
			(*highlighted[nh])[2] += 10;
		} else {
			for (int c = 0; c < 3; c++) {
				(*highlighted[nh])[c] /=2;
			}
		}
	}
	numhighlighted = 0;
}

//highlights the pixel around the cursor, as well as the corner pixels in non-closeups
void Area::setFocus(int areax, int areay)
{
	int x, y;
	activex = areax;
	activey = areay;
	DeepPixel* cen = &trace->deepBuffer[areay][areax];
	//highlight the pixels around the corners of x,y
	if (!closeup) {
		if (cen->numCorners > 0) {
			for (int c = 0; c < trace->deepBuffer[areay][areax].numCorners; c++) {
				y = cen->corners[c]->row;
				x = cen->corners[c]->col;
				for (int i = -1; i < 2; i++) {
					for (int j = -1; j < 2; j++) {
						if (j == 0 && i == 0) continue;
						if (y+j >= 0 && y+j < size && x+i >= 0 && x+i < size) {
							bool alreadyThere = false;
							for (int nh = 0; nh < numhighlighted; nh++) {
								if (highlighted[nh] == &area[I(y+j,x+i)]) {
									alreadyThere = true;
								}
							}
							if (y+j == cen->row && x+i == cen->col) {
								alreadyThere = true;
							}
							for (int corn = 0; corn < trace->deepBuffer[areay][areax].numCorners; corn++) {
								if (y+j == cen->corners[corn]->row && x+i == cen->corners[corn]->col) {
									alreadyThere = true;
								}
							}
							if (!alreadyThere) {	
								highlighted[numhighlighted++] = &area[I(y+j,x+i)];
							}
						}
					}
				}
			}
		}
		x = activex;
		y = activey;
	} else { //yes a closeup
		int h = Container::findLight(activey, activex);
		if (h >= 0) {
			activeLight = h;
			Container::direction2xy(x, y, thelights[h].direction, size);
		} else {
			x = 0;
			y = 0;
		}
	}
	//highlight the 8 pixels around x,y
	for (int i = -1; i < 2; i++) {
		for (int j = -1; j < 2; j++) {
			if (j == 0 && i == 0) continue;
			if (y+j >= 0 && y+j < size && x+i >= 0 && x+i < size) {
				bool alreadyThere = false;
				for (int nh = 0; nh < numhighlighted; nh++) {
					if (highlighted[nh] == &area[I(y+j,x+i)]) {
						alreadyThere = true;
					}
				}
				if (!closeup && (cen->numCorners > 0)) {
					for (int corn = 0; corn < trace->deepBuffer[areay][areax].numCorners; corn++) {
						if (y+j == cen->corners[corn]->row && x+i == cen->corners[corn]->col) {
							alreadyThere = true;
						}
					}
				}
				if (!alreadyThere) {	
					highlighted[numhighlighted++] = &area[I(y+j,x+i)];
				}
			}
		}
	}
	for (int nh = 0; nh < numhighlighted; nh++) {
		if (closeup) {
			(*highlighted[nh])[0] += 10;
			(*highlighted[nh])[1] -= 10;
			(*highlighted[nh])[2] -= 10;
		} else if (!highlightyellow) {
			P3D_VScaleEq((*highlighted[nh]), 2);
		} else {
			(*highlighted[nh])[0] += 10;
			(*highlighted[nh])[1] += 10;
			(*highlighted[nh])[2] -= 10;
		}
	}
}

//called from passivemotion on all the areas
void Area::setParent(Area* parea)
{
	if (closeup) {
		if (parea->closeup) {
			return;
		}
	} else { //this is not a closeup
		if (type == ONELIGHT && parea->closeup) {
			if (lightused == Container::findLight(parea->activey, parea->activex)) {
				return;
			}
		} else {
			return;
		}
	}
	//by now we're sure that we'll actually have to do some work
	parentArea = parea;
	fillArea();
}

void Area::fillArea() {
	unFocus();
	int row, col;
	float divisor;
	int thelight;
	DeepPixel* dp;
	if (closeup) {
		if (parentArea !=  NULL) {
			dp = &parentArea->trace->deepBuffer[parentArea->activey][parentArea->activex];
			if (corner != 0 && (dp->numCorners > 0)) {
				dp = parentArea->trace->deepBuffer[parentArea->activey][parentArea->activex].corners[corner-1];
			}
		} else {
			return;
		}
		if (dp->prediction == NULL) {
			return;
		}
	}
	if (normalize) {
		divisor = trace->numlights;
	} else {
		divisor = 255.0;
	}
	switch (type) {
	case TRACED:
		for (row = 0; row < size; row++) {
			for (col = 0; col < size; col++) {
				P3D_Assign_Vector(area[I(row,col)], trace->traced[row][col]);
			}
		}
		break;
	case DIFF:
		break;
	case NUMVISIBLE:
		for (row = 0; row < size; row++) {
			for (col = 0; col < size; col++) {
				int trackvis = 0;
				if (trace->deepBuffer[row][col].prediction != NULL) {
					for (int h = 0; h < trace->numlights; h++) {
						if (trace->deepBuffer[row][col].prediction[h] == PREDVIS ||
							trace->deepBuffer[row][col].prediction[h] == WASTRACEDVIS) {
							trackvis++;
						}
					}
				}
				P3D_Make_Gray_Vector(area[I(row,col)], trackvis / divisor);
			}
		}
		break;
	case WORKDONE:
		for (row = 0; row < size; row++) {
			for (col = 0; col < size; col++) {
				int trackwork = 0;
				if (trace->deepBuffer[row][col].prediction != NULL) {
					for (int h = 0; h < trace->numlights; h++) {
						if (trace->deepBuffer[row][col].prediction[h] == WASTRACEDVIS ||
							trace->deepBuffer[row][col].prediction[h] == WASTRACEDINVIS) {
							trackwork++;
						}
					}
				}
				P3D_Make_Gray_Vector(area[I(row,col)], trackwork / divisor);
			}
		}
		break;
	case ERRORS:
		for (row = 0; row < size; row++) {
			for (col = 0; col < size; col++) {
				int trackvis = 0;
				int trackinvis = 0;
				if (trace->deepBuffer[row][col].prediction != NULL) {
					for (int h = 0; h < trace->numlights; h++) {
						if (trace->deepBuffer[row][col].prediction[h] == PREDVIS &&
							trace->deepBuffer[row][col].extrainfo[h] == INCORRECT) {
							trackvis++;
						}
						if (trace->deepBuffer[row][col].prediction[h] == PREDINVIS &&
							trace->deepBuffer[row][col].extrainfo[h] == INCORRECT) {
							trackinvis++;
						}
					}
				}
				P3D_Make_Vector(area[I(row,col)], trackinvis / divisor, 0, trackvis / divisor);
			}
		}
		break;
	case ONELIGHT:
		if (parentArea == NULL) break;
		if (parentArea->parentArea == NULL) break;
		thelight = Container::findLight(parentArea->activey, parentArea->activex);
		lightused = thelight;
		if (thelight < 0) break;
		for (row = 0; row < size; row++) {
			for (col = 0; col < size; col++) {
				dp = &(parentArea->parentArea->trace->deepBuffer[row][col]);
				if (dp->objectID != 0) {
					if (dp->extrainfo[thelight] == INCORRECT) {
						if (dp->prediction[thelight] == PREDINVIS) {
							P3D_Make_Vector(area[I(row,col)], 1, 0, 0);
							continue;
						}
						if (dp->prediction[thelight] == PREDVIS) {
							P3D_Make_Vector(area[I(row,col)], 0, 0, 1);
							continue;
						}
					}
					if(parentArea->parentArea->trace->deepBuffer[row][col].prediction[thelight] == PREDVIS ||
						parentArea->parentArea->trace->deepBuffer[row][col].prediction[thelight] == WASTRACEDVIS) {
						float ndotl;
						P3D_VDot(ndotl, parentArea->parentArea->trace->deepBuffer[row][col].normal, thelights[thelight].direction);
						if (normalize) {
							P3D_Make_Vector(area[I(row,col)], ndotl*thelights[thelight].color[0]*parentArea->parentArea->trace->numlights,
														ndotl*thelights[thelight].color[1]*parentArea->parentArea->trace->numlights,
														ndotl*thelights[thelight].color[2]*parentArea->parentArea->trace->numlights);
						} else {
							P3D_Make_Vector(area[I(row,col)], ndotl, ndotl, ndotl);
						}
					}
					else {
						P3D_Make_Vector(area[I(row,col)], 0, 0, 0);
					}
				} else {
					P3D_Make_Vector(area[I(row,col)], 1, 0, 0);
				}
			}
		}
		break;
//the closeups
	case ALLLIGHTS:
		if (parentArea == NULL) break;
		for (row = 0; row < size; row++) {
			for (col = 0; col < size; col++) {
				thelight = Container::findLight(row, col);
				if (thelight < 0) {
					P3D_Make_Vector(area[I(row,col)], 0, 0, 0);
					continue;
				}
				P3D_Make_Vector(area[I(row,col)],
					parentArea->trace->predinviserrors[thelight]/(float)trace->numlights,
					0,
					parentArea->trace->predviserrors[thelight]/(float)trace->numlights);
				if (col < (size-1) && row < (size-1)) {
					if (thelight != Container::findLight(row, col+1)) {
						P3D_Make_Vector(area[I(row,col)], 0, 0, 0);
					}
					if (thelight != Container::findLight(row+1, col)) {
						P3D_Make_Vector(area[I(row,col)], 0, 0, 0);
					}
				}
			}
		}
		break;
	case PREDICTION:
		for (row = 0; row < size; row++) {
			for (col = 0; col < size; col++) {
				thelight = Container::findLight(row, col);
				if (thelight < 0) {
					P3D_Make_Vector(area[I(row,col)], 0, 0, 0);
					continue;
				}
				if (dp->extrainfo[thelight] == CORRECT && dp->prediction[thelight] == WASTRACEDVIS) {
					P3D_Make_Vector(area[I(row,col)], 1.0, 1.0, 1.0);
				} else if (dp->extrainfo[thelight] == CORRECT && dp->prediction[thelight] == WASTRACEDINVIS) {
					P3D_Make_Vector(area[I(row,col)], 1.0, 1.0, 0.0);
				} else if (dp->extrainfo[thelight] == INCORRECT &&
					(dp->prediction[thelight] == WASTRACEDVIS || dp->prediction[thelight] == WASTRACEDINVIS)) {
					P3D_Make_Vector(area[I(row,col)], 0.0, 0.0, 0.0);
				} else {
					switch(dp->prediction[thelight]) {
					case BHORIZON:  //below horizon - blue
	   				   P3D_Make_Vector(area[I(row,col)], 0, 0, 1);
					   break;
					case WASTRACEDVIS: //unpredictable - light grey
						P3D_Make_Vector(area[I(row,col)], 0.7, 0.7, 0.7);
						break;
					case WASTRACEDINVIS: //unpredictable - dark grey
						P3D_Make_Vector(area[I(row,col)], 0.3, 0.3, 0.3);
						break;
					case PREDVIS:
						if (dp->extrainfo[thelight] == CORRECT) {
							P3D_Make_Vector(area[I(row,col)], 0.0, 0.7, 0.0);
						} else {
							P3D_Make_Vector(area[I(row,col)], 1.0, 0.0, 0.0);
						}
						break;
					case PREDINVIS:
						if (dp->extrainfo[thelight] == CORRECT) {
							P3D_Make_Vector(area[I(row,col)], 0.0, 0.3, 0.0);
						} else {
							P3D_Make_Vector(area[I(row,col)], 0.6, 0.0, 0.0);
						}
						break;
				   }
				}
				if (col < (size-1) && row < (size-1)) {
					if (thelight != Container::findLight(row, col+1)) {
						P3D_Make_Vector(area[I(row,col)], 0, 0, 0);
					}
					if (thelight != Container::findLight(row+1, col)) {
						P3D_Make_Vector(area[I(row,col)], 0, 0, 0);
					}
				}
			}
		}
		break;
	}
}

void Area::setType(AreaType thetype)
{
	if (type == thetype) {
		return;
	}
	type = thetype;
	parentArea = NULL;
	highlightyellow = true;
	switch(thetype) {
	case TRACED:
		highlightyellow = false;
	case DIFF:
	case NUMVISIBLE:
	case WORKDONE:
	case ERRORS:
	case ONELIGHT:
		closeup = false;
		break;
	case PREDICTION:
	case ALLLIGHTS:
		closeup = true;
		break;
	}
	for (int a = 0; a < 6; a++) {
		if (Container::areas[a]->parentArea == this) {
			Container::areas[a]->setParent(this);
		}
	}
}

void Area::brighten(float amount)
{
	for (int row = 0; row < size; row++) {
		for (int col = 0; col < size; col++) {
			P3D_VScaleEq(area[I(row,col)], amount);
		}
	}

}
