// Aner Ben-Artzi - Efficient Coherence-Based Visibility Sampling (CBVS) of Environment Maps
// Columbia University 2004
//
// VVTrace.cpp: implementation of the VVTrace class.
//
//////////////////////////////////////////////////////////////////////

/*
The colors are calculated based on the predicted visibility.

However, the deepBuffer is filled with correct information, always,
so that errors do not propogate.
*/

#include "VVTrace.h"

char* VVTrace::flagnames[NUMFLAGS] = {
	"1bak",
	"grid",
	"3Obj",
	"bdry",
	"hor",
	"unc",
	"warp",
	"hi-C",
	"opt",
	"strng",
	"t",
	"v",
	"r"};

	bool VVTrace::globalFlags[];

/*
Constructor - creats an object that can trace the scene stored in Image_Info.
The traced image has resolution thesize X thesize.
The traced image uses thenumlights lights in the scene.

Assumptions:
size is a power of 2
stepSize >= 1
numlights >= 1
*/
VVTrace::VVTrace() {
	className = "VVTrace";
};


void VVTrace::create(int thesize, int thenumlights)
{
	size = thesize;
	for (int flag = 0; flag < NUMFLAGS; flag++) {
		flags[flag] = globalFlags[flag];
	}

	if (flags[USE_ONEBACK] && flags[USE_GRID]) {
		printf("warning: oneback and grid\n");
	}

	numlights = thenumlights;

	numTraced = 0;
	numAboveHorizon = 0;
	numVisCor = 0;
	numInvisCor = 0;
	numVisIncor = 0;
	numInvisIncor = 0;
	numTracedVis = 0;
	numTracedInvis = 0;
	numBHorizon = 0;
	numPixelsHit = 0;
	numFloodChange = 0;
	numFloodSame = 0;
	numNotEnoughNeighbors = 0;
	completed = false;

	traced = new Point3d*[size];
	deepBuffer = new DeepPixel*[size];
	for (int i = 0; i < size; i++) {
		traced[i] = new Point3d[size];
		deepBuffer[i] = new DeepPixel[size];
		for (int j = 0; j < size; j++) {
			deepBuffer[i][j].prediction = NULL;
			deepBuffer[i][j].extrainfo = NULL;
			deepBuffer[i][j].occluders = NULL;
		}
	}
	visible = new bool[numlights];
	cblocked = new int[numlights];
	bestfall = new float[numlights];
	bestoccluder = new Blocker_List*[numlights];
	usedNeighbors = new int[numlights];

#ifndef RAYTRACER
	setTitle();
	predviserrors = new int[numlights];
	predinviserrors = new int[numlights];
	for (int h = 0; h < numlights; h++) {
		predviserrors[h] = 0;
		predinviserrors[h] = 0;
	}
#endif //RAYTRACER
}

/*Standard destructor*/
VVTrace::~VVTrace()
{
	printf("Deleting Trace.\n");
	for (int i = 0; i < size; i++) {
		delete[] traced[i];
		for (int j = 0; j < size; j++) {
			delete[] Image_Info[i][j].depths;
			delete[] deepBuffer[i][j].prediction;
			delete[] deepBuffer[i][j].extrainfo;
			Blocker_List* runner = deepBuffer[i][j].occluders;
			Blocker_List* ahead;
			while (runner != NULL) {
				ahead = runner->next;
				delete runner;
				runner = ahead;
			}
		}
		delete[] deepBuffer[i];
	}
	delete[] traced;
	delete[] cblocked;
	delete[] deepBuffer;
	delete[] visible;
	delete[] bestfall;
	delete[] bestoccluder;
	delete[] usedNeighbors;
#ifndef RAYTRACER
	delete[] predviserrors;
	delete[] predinviserrors;
#endif //RAYTRACER

}


void VVTrace::tracePixel(int row, int col) {
	crow = row;
	ccol = col;
	tracePixel();
};

/*
Performs the steps for tracing the pixel at crow,ccol.
First light visibility is determined.  Then the BRDF is used to add actual color
*/
//numPixelsHit++
void VVTrace::tracePixel()
{
	if (deepBuffer[crow][ccol].objectID != 0) {
		numPixelsHit++;
		if (!sameObject()) {
			deepBuffer[crow][ccol].numCorners = 0;
		}
		if (flags[USE_TRUE]) {
			deepBuffer[crow][ccol].numCorners = 0;
		}
		if (flags[USE_WARP] || flags[USE_ALWAYS_VISIBLE] || flags[USE_TRUE]) {
			predictAllLights(); //after this, visible[] contains the boolean vector of light visibility
		} else {
			predictAllLightsNoWarp();
		}
		traceUnpredictable();
#ifndef RAYTRACER
		fillExtraInfo();
#endif //RAYTRACER

		if (!notEnoughCorners(deepBuffer[crow][ccol].numCorners)) {
			if (flags[USE_UFLOODING] && !flags[USE_BFLOODING]) {
				floodGray();
			} 
			if (flags[USE_BFLOODING] && !flags[USE_UFLOODING]) {
				floodBoundry();
			}
			if (flags[USE_BFLOODING] && flags[USE_UFLOODING]) {
				floodIntersection();
			}
			if (flags[USE_HFLOODING] || (flags[USE_HIGHC] && highCurvature(crow, ccol))) {
				floodHorizon();
			}
			if (flags[USE_STRONG]) {
				traceStrongestSources();
				traceUnpredictable();
			}
		}
	} else {
		for (int h = 0; h < numlights; h++) {
			visible[h] = false;
		}
	}
	Point3d brdfResult;
	Point3d eyeRay;
	for (clight = 0; clight < numlights; clight++) {
		if (visible[clight]) {
#ifdef RAYTRACER
			doBRDF(brdfResult, clight, deepBuffer[crow][ccol].normal, deepBuffer[crow][ccol].objectID);
#else //not RAYTRACER
			doBRDF(brdfResult, DIFFUSE, clight, eyeRay, deepBuffer[crow][ccol].normal, deepBuffer[crow][ccol].objectID);
#endif //RAYTRACER
			P3D_VAddEq(traced[crow][ccol], brdfResult);
		}
	}
}

/*
Trace every nth pixel (in both dimentions).  n = n/2 at each pass.
The pattern always places new traces in the center of 4 existing pixels.
*/
void VVTrace::traceAlternatingGrid()
{
	cstep = STEPSIZE;
	//initial gird points
	for (crow = 0; crow <= (size-1); crow += STEPSIZE) {
		for (ccol = 0; ccol <= (size-1); ccol += STEPSIZE) {
			deepBuffer[crow][ccol].numCorners = 0;
			tracePixel();
		}
	}
	//alternating grid points
	for (cstep = STEPSIZE; cstep > 1; cstep /= 2) {
		bool oldWarp = flags[USE_WARP];
		if (cstep == 2 && flags[USE_OPTIMIZE]) {
			flags[USE_WARP] = false;
			flags[USE_3OBJECT] = true;
			flags[USE_BFLOODING] = false;
			flags[USE_UFLOODING] = false;
		}
		printf("Step: \n%d", cstep); 
		//diagonal
		for (crow = 0 + cstep/2; crow < size; crow += cstep) {
			printf(".");
			for (ccol = 0 + cstep/2; ccol < size-1; ccol += cstep) {
				findGridCorners(crow, ccol);
				tracePixel();
			}
		}
		//cross1
		for (crow = 0 + cstep/2; crow < size; crow += cstep) {
			printf(".");
			for (ccol = 0; ccol <= size-1; ccol+= cstep) {
				findGridCorners(crow, ccol);
				tracePixel();
			}
		}
		//cross2
		for (crow = 0; crow < size; crow += cstep) {
			printf(".");
			for (ccol = 0 + cstep/2; ccol < size-1; ccol += cstep) {
				findGridCorners(crow, ccol);
				tracePixel();
			}
		}
		flags[USE_WARP] = oldWarp;
	}
}

void VVTrace::findGridCorners(int row, int col) {
	bool edge;
	if (row == 0 || col == 0 ||
		row == size-1 || col == size-1) {
		edge = true;
	}
	if (row == 0 && col == 0 ||
		row == size-1 && col == size-1 ||
		row == 0 && col == size-1 ||
		row == size-1 && col == 0 ||
		(row == (size-1)/2 && col == (size-1)/2)) {
			deepBuffer[row][col].numCorners = 0;
			return; //just for safety
	}
	int rowpot, colpot, step;
	if (row != 0) {
		rowpot = 1;
		while (row % rowpot == 0) {
			rowpot *= 2;
		}
		rowpot /= 2;
		step = rowpot*2;
	}
	if (col != 0) {
		colpot = 1;
		while (col % colpot == 0) {
			colpot *= 2;
		}
		colpot /= 2;
		if (rowpot > colpot || row == 0) {
			step = colpot*2;
		}
	}
	if (step/2 >= STEPSIZE) {
		deepBuffer[row][col].numCorners = 0;
		return;
	}
	if ((row != 0) && (col != 0) && (rowpot == colpot)) { //diagonal
		deepBuffer[row][col].numCorners = 4;
		deepBuffer[row][col].corners[0] = &deepBuffer[row-step/2][col-step/2];
		deepBuffer[row][col].corners[1] = &deepBuffer[row-step/2][col+step/2];
		deepBuffer[row][col].corners[2] = &deepBuffer[row+step/2][col-step/2];
		deepBuffer[row][col].corners[3] = &deepBuffer[row+step/2][col+step/2];
	} else {
		if (col == 0) {
			deepBuffer[row][col].numCorners = 3;
			deepBuffer[row][col].corners[0] = &deepBuffer[row][col+step/2];
			deepBuffer[row][col].corners[1] = &deepBuffer[row-step/2][col];
			deepBuffer[row][col].corners[2] = &deepBuffer[row+step/2][col];
		} else if (row == 0) {
			deepBuffer[row][col].numCorners = 3;
			deepBuffer[row][col].corners[0] = &deepBuffer[row][col-step/2];
			deepBuffer[row][col].corners[1] = &deepBuffer[row][col+step/2];
			deepBuffer[row][col].corners[2] = &deepBuffer[row+step/2][col];
		} else if (col == size-1) {
			deepBuffer[row][col].numCorners = 3;
			deepBuffer[row][col].corners[0] = &deepBuffer[row][col-step/2];
			deepBuffer[row][col].corners[1] = &deepBuffer[row-step/2][col];
			deepBuffer[row][col].corners[2] = &deepBuffer[row+step/2][col];
		} else if (row == size-1) {
			deepBuffer[row][col].numCorners = 3;
			deepBuffer[row][col].corners[0] = &deepBuffer[row][col-step/2];
			deepBuffer[row][col].corners[1] = &deepBuffer[row][col+step/2];
			deepBuffer[row][col].corners[2] = &deepBuffer[row-step/2][col];
		} else {
			deepBuffer[row][col].numCorners = 4;
			deepBuffer[row][col].corners[0] = &deepBuffer[row][col-step/2];
			deepBuffer[row][col].corners[1] = &deepBuffer[row][col+step/2];
			deepBuffer[row][col].corners[2] = &deepBuffer[row-step/2][col];
			deepBuffer[row][col].corners[3] = &deepBuffer[row+step/2][col];
		}
	}
	for (int c = 0; c < deepBuffer[row][col].numCorners; c++) {
		if (deepBuffer[row][col].corners[c]->objectID == 0) {
			if (deepBuffer[row][col].numCorners == 4) {  //we allow one of the corners to be untraced (just like borders)
				deepBuffer[row][col].numCorners = 3;
				deepBuffer[row][col].corners[c] = deepBuffer[row][col].corners[3];
				c--;
			} else {
				deepBuffer[row][col].numCorners = 0;
				return;
			}
		}
	}
//	deepBuffer[row][col].usesCorners = numCornersGlobal;
}

void VVTrace::traceUnpredictable() {
	DeepPixel& dp = deepBuffer[crow][ccol];
	Point3d occluder;
	for (clight = 0; clight < numlights; clight++) {		
		if (dp.prediction[clight] == TOBETRACED) {
			visible[clight] = traceShadowRay(occluder);
			if (visible[clight]) {
				dp.prediction[clight] = WASTRACEDVIS;
			} else {
				dp.prediction[clight] = WASTRACEDINVIS;
				if (flags[USE_WARP]) {
					addOccluder(crow, ccol, clight, occluder);
				}
			}
		}
	}
}

void VVTrace::predictAllLightsNoWarp() {
	DeepPixel& dp = deepBuffer[crow][ccol];

	//mark below the Horizon
	float ndotl;
	for (clight = 0; clight < numlights; clight++) {
		P3D_VDot(ndotl, thelights[clight].direction, dp.normal);
		if (ndotl <= 0) {
			dp.prediction[clight] = BHORIZON;
			visible[clight] = false;
		} else {
			numAboveHorizon++;
			dp.prediction[clight] = PREDVIS;
		}
	}
	
	//trace all above horizon if NOT ENOUGH CORNERS
	if (notEnoughCorners(dp.numCorners)) {
		for (clight = 0; clight < numlights; clight++) {
			if (dp.prediction[clight] != BHORIZON) {
				numNotEnoughNeighbors++;
				dp.prediction[clight] = TOBETRACED;
			}
		}
		return;
	}

	//Attempt to predict by unanimous vote
	int count;
	char pred;
	for (clight = 0; clight < numlights; clight++) {
		if (dp.prediction[clight] == BHORIZON) {
			continue;
		}
		count = 0;
		for (ccorner = 0; ccorner < dp.numCorners; ccorner++) {
			pred = dp.corners[ccorner]->prediction[clight];
			if (pred == WASTRACEDINVIS ||
				pred == PREDINVIS ||
				pred == BHORIZON) {
				count++;
			}
		}
		if (count == dp.numCorners) {
			dp.prediction[clight] = PREDINVIS;
			visible[clight] = false;
		} else if (count == 0) {
			dp.prediction[clight] = PREDVIS;
			visible[clight] = true;
		} else {
			dp.prediction[clight] = TOBETRACED;
		}
	}
}


//any time a blockage is found, it is added to the cblocked
void VVTrace::predictAllLights() {
	DeepPixel& dp = deepBuffer[crow][ccol];
	int numcorners = dp.numCorners;

	//check for ndotl <= 0
	float ndotl;
	for (clight = 0; clight < numlights; clight++) {
		cblocked[clight] = 0;									//blockage is low
		bestfall[clight] = -1;									//any occluder is closer
		bestoccluder[clight] = NULL;							//no occluder found yet
		usedNeighbors[clight] = -1;								//no neighbors yet checked if they are affected
		graph[clight]->nextSibling = NULL;							//toBeChecked list is empty

		P3D_VDot(ndotl, thelights[clight].direction, dp.normal);
		if (ndotl <= 0) {
			dp.prediction[clight] = BHORIZON;
			visible[clight] = false;
		} else {
			numAboveHorizon++;
			dp.prediction[clight] = PREDVIS;
			visible[clight] = true;
		}
	}
	if (flags[USE_ALWAYS_VISIBLE]) {
		return;
	}

	//check for sufficient corners
	if (notEnoughCorners(dp.numCorners)) {
		for (clight = 0; clight < numlights; clight++) {
			if (dp.prediction[clight] != BHORIZON) {
				numNotEnoughNeighbors++;
				dp.prediction[clight] = TOBETRACED;
			}
		}
		return;
	}

		
	//add to certainties from blockages at corners
	Blocker_List* walker;
	for (ccorner = 0; ccorner < numcorners; ccorner++) {
		walker = dp.corners[ccorner]->occluders;
		while (walker != NULL) {
			addSingleBlockage(walker); //takes care of setting [bestoccluder]s
			walker = walker->next;
		}
	}
	//check if ray must be traced, set depth and prediction
	for (clight = 0; clight < numlights; clight++) {
		if (dp.prediction[clight] != BHORIZON){
			if (cblocked[clight] < numcorners) { //blockage certainty is low
				if (cblocked[clight] > 0) { //visibility certainty is also low
					dp.prediction[clight] = TOBETRACED;
				} else { // visible with high confidence
					visible[clight] = true;
					deepBuffer[crow][ccol].prediction[clight] = PREDVIS;
				}
			} else { //blocked with high confidence
				visible[clight] = false;
				deepBuffer[crow][ccol].prediction[clight] = PREDINVIS;
				addOccluder(crow, ccol, clight, bestoccluder[clight]->occluder);
			}
		}
	}
}



#ifndef RAYTRACER

//numTracedVis++;
//numTracedInvis++;
//numBHorizon++;
//numVisCor++;
//numVisIncor++;
//predviserrors[clight]++;
//numInvisIncor++;
//predinviserrors[clight]++;
//numInvisCor++;
void VVTrace::fillExtraInfo() {
	bool actually_visible;
	Point3d occluder;
	DeepPixel& dp = deepBuffer[crow][ccol];
	int il2 = Container::intlog2(cstep);
	for (clight = 0; clight < numlights; clight++) {
		actually_visible = traceShadowRay(occluder);
		switch (dp.prediction[clight]) {
		case WASTRACEDVIS:
			dp.extrainfo[clight] = CALCULATED;
			numTracedVis++;
			break;
		case WASTRACEDINVIS:
			dp.extrainfo[clight] = CALCULATED;
			numTracedInvis++;
			break;
		case BHORIZON:
			dp.extrainfo[clight] = CALCULATED;
			numBHorizon++;
			break;
		case PREDVIS:
			if (actually_visible) {
				dp.extrainfo[clight] = CORRECT;
				numVisCor++;
			} else {
				dp.extrainfo[clight] = INCORRECT;
				numVisIncor++;
				predviserrors[clight]++;
				if (flags[USE_REAL]) {
					addOccluder(crow, ccol, clight, occluder);
				}
			}
			break;
		case PREDINVIS:
			if (actually_visible) {
				dp.extrainfo[clight] = INCORRECT;
				numInvisIncor++;
				predinviserrors[clight]++;
				if (flags[USE_REAL]) {
					removeOccluder(crow, ccol, clight);
				}
			} else {
				dp.extrainfo[clight] = CORRECT;
				numInvisCor++;
			}
			break;
		}
	}
}
#endif //RAYTRACER

#ifdef RAYTRACER
extern bool Intersection (INTERSECTION *Ray_Intersection, OBJECT *Object, RAY *Ray);
extern void Initialize_Ray_Containers(RAY *Ray, unsigned long optimisiation_flags);
extern int anerfunc(RAY *ray, DBL x, DBL y, int ray_number);
#endif //YES RAYTRACER

//returns true if light is visible
bool VVTrace::traceShadowRay(Point3d& result, int row, int col, int light) {
	float depth;
	numTraced++;

#ifdef RAYTRACER
	Image_Info[row][col].depths[light] = VISIBLE;

	RAY Shadow_Ray;
	Shadow_Ray.Initial[0] = deepBuffer[row][col].intersection[0];
	Shadow_Ray.Initial[1] = deepBuffer[row][col].intersection[1];
	Shadow_Ray.Initial[2] = deepBuffer[row][col].intersection[2];
	Shadow_Ray.Direction[0] = thelights[light].direction[0];
	Shadow_Ray.Direction[1] = thelights[light].direction[1];
	Shadow_Ray.Direction[2] = thelights[light].direction[2];
	Shadow_Ray.Index = 0;
	Initialize_Ray_Containers(&Shadow_Ray, OPTIMISE_NONE);

	INTERSECTION Best_Intersection, New_Intersection;
	OBJECT * Object;
	bool Intersection_Found = false;
	Best_Intersection.Depth = BOUND_HUGE;
	Best_Intersection.Object = NULL;

	for (Object = Frame.Objects; Object != NULL; Object = Object -> Sibling)
	{
	  if (Object->Type != 37) ///don't do lights type = 37
	  {
		  if (Intersection(&New_Intersection, Object, &Shadow_Ray))
		  {
			if (New_Intersection.Depth < Best_Intersection.Depth)
			{
			  Best_Intersection = New_Intersection;

			  Intersection_Found = true;
			  if (!flags[USE_WARP]) {
				  break;
			  }
			}
		  }
    
	  }
	}
	if (Intersection_Found) {
		Image_Info[row][col].depths[light] = Best_Intersection.Depth;
	}
#endif //YES RAYTRACER

	depth = Image_Info[row][col].depths[light];
	P3D_VScale(result, thelights[light].direction, depth);
	P3D_VAddEq(result, deepBuffer[row][col].intersection);
	return depth == VISIBLE;
}


bool VVTrace::traceShadowRay(Point3d& result) {
	return traceShadowRay(result, crow, ccol, clight);
}

bool VVTrace::traceShadowRay()
{
	Point3d temp;
	return traceShadowRay(temp, crow, ccol, clight);
}

bool VVTrace::traceShadowRay(int row, int col, int light)
{
	Point3d temp;
	return traceShadowRay(temp, row, col, light);
}


void VVTrace::tracePrimaryRay()
{
	deepBuffer[crow][ccol].row = crow;
	deepBuffer[crow][ccol].col = ccol;
	for (int c = 0; c < 4; c++) {
		deepBuffer[crow][ccol].corners[c] = NULL;
	}
	deepBuffer[crow][ccol].numCorners = 0;
	deepBuffer[crow][ccol].occluders = NULL;


#ifdef RAYTRACER
	Image_Info[crow][ccol].objectID = 0;

	RAY Camera_Ray;
    if (anerfunc(&Camera_Ray, (DBL)ccol, (DBL)crow, 0)) {
		INTERSECTION Best_Intersection, New_Intersection;
		OBJECT * Object;
		bool Intersection_Found = false;
		Best_Intersection.Depth = BOUND_HUGE;
		Best_Intersection.Object = NULL;

		for (Object = Frame.Objects; Object != NULL; Object = Object -> Sibling)
		{
		  if (Object->Type != 37) ///don't do lights type = 37
		  {
			  if (Intersection(&New_Intersection, Object, &Camera_Ray))
			  {
				if (New_Intersection.Depth < Best_Intersection.Depth)
				{
				  Best_Intersection = New_Intersection;

				  Intersection_Found = true;
				}
			  }
        
		  }
		}
		if (Intersection_Found) {
			Image_Info[crow][ccol].objectID = (int)Best_Intersection.Object;
			Image_Info[crow][ccol].iPoint[0] = Best_Intersection.IPoint[0];
			Image_Info[crow][ccol].iPoint[1] = Best_Intersection.IPoint[1];
			Image_Info[crow][ccol].iPoint[2] = Best_Intersection.IPoint[2];

			VECTOR Raw_Normal;
			DBL Normal_Direction;
			Normal(Raw_Normal, Best_Intersection.Object, &Best_Intersection);
			VDot(Normal_Direction, Raw_Normal, Camera_Ray.Direction);
			if (Normal_Direction > 0.0)
			{
				VScaleEq(Raw_Normal, -1.0);
			}
			Image_Info[crow][ccol].iNormal[0] = Raw_Normal[0];
			Image_Info[crow][ccol].iNormal[1] = Raw_Normal[1];
			Image_Info[crow][ccol].iNormal[2] = Raw_Normal[2];
			Image_Info[crow][ccol].depths = new float[numlights];
		}
	}
#endif //YES RAYTRACER



	deepBuffer[crow][ccol].objectID = Image_Info[crow][ccol].objectID;
	if (deepBuffer[crow][ccol].objectID != 0) {
		P3D_Assign_Vector(deepBuffer[crow][ccol].intersection, Image_Info[crow][ccol].iPoint);
		P3D_Assign_Vector(deepBuffer[crow][ccol].normal, Image_Info[crow][ccol].iNormal);
		deepBuffer[crow][ccol].prediction = new char[numlights];
		deepBuffer[crow][ccol].extrainfo = new char[numlights];
	} else {
		P3D_Make_Vector(deepBuffer[crow][ccol].intersection, 0, 0, 0);
		P3D_Make_Vector(deepBuffer[crow][ccol].normal, 0, 0, 0);
		deepBuffer[crow][ccol].prediction = NULL;
		deepBuffer[crow][ccol].extrainfo = NULL;
	}
	P3D_Make_Vector(traced[crow][ccol], 0, 0, 0);
}

#ifdef RAYTRACER
void VVTrace::doBRDF(Point3d& result, int thelight, Point3d& theNormal, int objectID){
	float ndotl;
	P3D_VDot(ndotl, thelights[thelight].direction, theNormal);
	if (ndotl < 0) {
		ndotl = 0;
	}
	OBJECT* Object = (OBJECT*)(objectID);
	P3D_Make_Vector(result, thelights[thelight].color[0] * ndotl * Object->Texture->Pigment->Colour[0],
		thelights[thelight].color[1] * ndotl * Object->Texture->Pigment->Colour[1],
		thelights[thelight].color[2] * ndotl * Object->Texture->Pigment->Colour[2]);

}
#endif


#ifndef RAYTRACER

void VVTrace::doBRDF(Point3d& result, BRDF thebrdf, int thelight, Point3d& theEyeRay, Point3d& theNormal, int objectID){
	float ndotl;
	P3D_VDot(ndotl, thelights[thelight].direction, theNormal);
	if (ndotl < 0) {
		ndotl = 0;
	}
	float m =  ndotl* ndotl;
	float intensity;
	switch (thebrdf) {
	case DIFFUSE:
		P3D_Make_Vector(result, thelights[thelight].color[0] * ndotl * Container::objectColor[objectID][0],
			thelights[thelight].color[1] * ndotl * Container::objectColor[objectID][1],
			thelights[thelight].color[2] * ndotl * Container::objectColor[objectID][2]);
		break;
	case BINARY:
		P3D_Make_Vector(result, thelights[thelight].color[0] * Container::objectColor[objectID][0] * m,
			thelights[thelight].color[1] * Container::objectColor[objectID][1] * m,
			thelights[thelight].color[2] * Container::objectColor[objectID][2] * m);
		break;
	case EXPANDED:
		intensity = thelights[thelight].color[0] +
			thelights[thelight].color[1] +
			thelights[thelight].color[2];
		m = intensity*intensity*(Container::numberOfLights*5);
		P3D_Make_Vector(result, thelights[thelight].color[0] * ndotl * Container::objectColor[objectID][0] * m,
			thelights[thelight].color[1] * ndotl * Container::objectColor[objectID][1] * m,
			thelights[thelight].color[2] * ndotl * Container::objectColor[objectID][2] * m);
		break;
	}
}
#endif //RAYTRACER

void VVTrace::findScanCorners(int row, int col) {
	if (row == 0 || col == 0) {
		deepBuffer[row][col].numCorners = 0;
	} else {
		deepBuffer[row][col].numCorners = 3;
		deepBuffer[row][col].corners[0] = &deepBuffer[row][col-1];
		deepBuffer[row][col].corners[1] = &deepBuffer[row-1][col-1];
		deepBuffer[row][col].corners[2] = &deepBuffer[row-1][col];
		if (flags[USE_ONEBACK]) {
			deepBuffer[crow][ccol].numCorners = 1;
		}
		for (int c = 0; c < deepBuffer[row][col].numCorners; c++) {
			if (deepBuffer[row][col].corners[c]->objectID == 0) {
				deepBuffer[row][col].numCorners = 0;
				return;
			}
		}
	}
}

void VVTrace::traceScanline() {
	cstep = STEPSIZE;
	for (crow = 0; crow < size; crow++) {
		printf("row %d   %c", crow, crow == size-1 ? '\n' : '\r');
		for (ccol = 0; ccol < size; ccol++) {
			findScanCorners(crow, ccol);
			tracePixel();
		}
	}
}


/*
Needs to set completed = true;
*/
void VVTrace::trace()
{
	printf("Begin tracing primary rays\n");
	for (crow = 0; crow < size; crow++) {
		for (ccol = 0; ccol < size; ccol++) {
			tracePrimaryRay();	//fills in objectID, intersection-point, and normal
		}
	}

	printf("Begin Tracing Shadows\n");
	if (flags[USE_GRID]) {
		traceAlternatingGrid();
	} else {
		traceScanline();
	}

	completed = true;
	setTitle();
	printf("Trace Complete.\n");

#ifndef RAYTRACER
	//sanity check
	int numPred = numVisCor + numVisIncor + numInvisCor + numInvisIncor;
	int numTraced = numTracedVis + numTracedInvis;
	if (numPixelsHit*numlights != (numPred + numTraced + numBHorizon)) {
		printf("Error: insanity exists.\n");
	}
	int totalVisIncor = 0;
	int totalInvisIncor = 0;
	for (int h = 0; h < numlights; h++) {
		totalVisIncor += predinviserrors[h];
		totalInvisIncor += predviserrors[h];
	}
	if ((totalVisIncor + totalInvisIncor) != ((numInvisIncor + numVisIncor) - numFloodChange)) {
		printf("Error: there's insanity a brewin'\n");
	}
#endif //RAYTRACER

}



void VVTrace::setTitle() {
	int offset = 0;
	char* intitle = title;
	for (int f = 0; f < NUMFLAGS; f++) {
		offset = sprintf(intitle, "%s=%c ", flagnames[f], flags[f]? 'T' : 'F');
		intitle += offset;
	}
}

void VVTrace::subprint(FILE* fp, int num, float under, char* text) {
	float percent;
	if (under != 0) {
		percent = num / under;
	} else {
		percent = 0;
	}
	fprintf(fp, "%s: %d : %.2f\n", text, num, 100*percent);
}	

void VVTrace::print(FILE* fp)
{
	fprintf(fp, "%s\n", title);
	int numPred, numTraced, numShadowRays;
	numTraced = numTracedVis + numTracedInvis;
	numPred = numVisCor + numVisIncor + numInvisCor + numInvisIncor;
	numShadowRays = numTraced + numPred;

	subprint(fp,
	numShadowRays,
	numPixelsHit * numlights,
	"Above Horizon");

	subprint(fp,
	numNotEnoughNeighbors,
	numPixelsHit * numlights,
	"Not enough neighbors");

	subprint(fp,
	numPred,
	numShadowRays,
	"Predicted");

	subprint(fp,
	numTraced,
	numShadowRays,
	"Work(Traced)");

	subprint(fp,
	numVisIncor + numInvisIncor,
	numPred,
	"Errors");

	subprint(fp,
	numVisIncor,
	numVisIncor + numInvisIncor,
	"Predicted Visible Errors");

	subprint(fp,
	numInvisIncor,
	numVisIncor + numInvisIncor,
	"Predicted Blocked Errors");

	subprint(fp,
	numFloodSame + numFloodChange,
	numShadowRays,
	"Flood Work");

	subprint(fp,
	numFloodChange,
	numVisIncor + numInvisIncor,
	"Flood Change");

	subprint(fp,
	numFloodSame + numFloodChange + numTraced,
	numShadowRays,
	"Final Work");

	subprint(fp,
	(numVisIncor + numInvisIncor) - numFloodChange,
	numPred - (numFloodSame + numFloodChange),
	"Final Errors");

	fprintf(fp, "\n");

}


void VVTrace::floodIntersection() {
	Light_Graph* mightBeAffected = NULL;
	Light_Graph* affectedTail = NULL;
	Light_Graph* runner;
	char clightPred;
	for (int h = 0; h < numlights; h++) {
		graph[h]->nextSibling = NULL;
	}
	for (clight = 0; clight < numlights; clight++) { //add the traced's neighboors
		clightPred = deepBuffer[crow][ccol].prediction[clight];
		if (clightPred == WASTRACEDVIS || clightPred == WASTRACEDINVIS) {
			if (mightBeAffected == NULL) {
				mightBeAffected = graph[clight];
				affectedTail = mightBeAffected;
			}
			runner = graph[clight]->firstChild;
			while (runner != NULL) {
				if ((deepBuffer[crow][ccol].prediction[runner->lightNumber] == PREDVIS &&
					clightPred == WASTRACEDINVIS) ||
					(deepBuffer[crow][ccol].prediction[runner->lightNumber] == PREDINVIS &&
					clightPred == WASTRACEDVIS)) {

					if (graph[runner->lightNumber]->nextSibling == NULL && 
						graph[runner->lightNumber] != affectedTail) {
						affectedTail->nextSibling = graph[runner->lightNumber];
						affectedTail = graph[runner->lightNumber];
					}
				}
				runner = runner->nextSibling;
			}
		}
	}
	floodOutwards(mightBeAffected, affectedTail);
}


void VVTrace::floodGray() {
	Light_Graph* mightBeAffected = NULL;
	Light_Graph* affectedTail = NULL;
	Light_Graph* runner;
	for (int h = 0; h < numlights; h++) {
		graph[h]->nextSibling = NULL;
	}
	for (clight = 0; clight < numlights; clight++) { //add the traced's neighboors
		if (deepBuffer[crow][ccol].prediction[clight] == WASTRACEDVIS ||
			deepBuffer[crow][ccol].prediction[clight] == WASTRACEDINVIS) {
			if (mightBeAffected == NULL) {
				mightBeAffected = graph[clight];
				affectedTail = mightBeAffected;
			}
			runner = graph[clight]->firstChild;
			while (runner != NULL) {
				if (graph[runner->lightNumber]->nextSibling == NULL && 
					graph[runner->lightNumber] != affectedTail) {
					affectedTail->nextSibling = graph[runner->lightNumber];
					affectedTail = graph[runner->lightNumber];
				}
				runner = runner->nextSibling;
			}
		}
	}
	floodOutwards(mightBeAffected, affectedTail);
}

//numFloodChange++;
//predviserrors[clight]--;
//predinviserrors[clight]--;
//numFloodSame++
void VVTrace::floodOutwards(Light_Graph* mightBeAffected, Light_Graph* affectedTail) {
	bool actually_visible;
	Light_Graph* runner;
	bool changed;

	Point3d occluder;
	while (mightBeAffected != NULL) { //flood, and possibly flood outwards
		clight = mightBeAffected->lightNumber;
		char& thisPred = deepBuffer[crow][ccol].prediction[clight];
		if (thisPred == PREDVIS ||
			thisPred == PREDINVIS) {
				actually_visible = traceShadowRay(occluder);
				if (actually_visible) {
					changed = (thisPred == PREDINVIS);
					thisPred = WASTRACEDVIS;
				} else {
					changed = (thisPred == PREDVIS);
					thisPred = WASTRACEDINVIS;
					if (flags[USE_WARP]) {
						removeOccluder(crow, ccol, clight);
						addOccluder(crow, ccol, clight, occluder);
					}
				}
				visible[clight] = actually_visible;
				if (changed) {
#ifndef RAYTRACER
					numFloodChange++;
					if (thisPred == PREDVIS) {
						predviserrors[clight]--;
					} else {
						predinviserrors[clight]--;
					}
#endif //RAYTRACER
					runner = mightBeAffected->firstChild;
					while (runner != NULL) {
						if (graph[runner->lightNumber]->nextSibling == NULL && 
							graph[runner->lightNumber] != affectedTail) {
							affectedTail->nextSibling = graph[runner->lightNumber];
							affectedTail = graph[runner->lightNumber];
						}
						runner = runner->nextSibling;
					}
				} else {
#ifndef RAYTRACER
					numFloodSame++;
#endif //RAYTRACER
				}
		}
		mightBeAffected = mightBeAffected->nextSibling;
	}
}

void VVTrace::floodBoundry() {
	Light_Graph* mightBeAffected = NULL;
	Light_Graph* affectedTail = NULL;
	Light_Graph* runner;
	for (int h = 0; h < numlights; h++) {
		graph[h]->nextSibling = NULL;
	}
	for (clight = 0; clight < numlights; clight++) {
		bool vis, alsovis, invis, alsoinvis;
		char alsopred;
		char pred = deepBuffer[crow][ccol].prediction[clight];
		vis = (pred == WASTRACEDVIS || pred == PREDVIS);
		invis = (pred == WASTRACEDINVIS || pred == PREDINVIS);
		runner = graph[clight]->firstChild;
		while (runner != NULL) {
			alsopred = deepBuffer[crow][ccol].prediction[runner->lightNumber];
			alsovis = (alsopred == WASTRACEDVIS || alsopred == PREDVIS);
			alsoinvis = (alsopred == WASTRACEDINVIS || alsopred == PREDINVIS);
			if ((vis && alsoinvis) || (invis && alsovis)) {
				if (mightBeAffected == NULL) {
					mightBeAffected = graph[clight];
					affectedTail = mightBeAffected;
				}
				if (graph[clight]->nextSibling == NULL && graph[clight] != affectedTail) {
					affectedTail->nextSibling = graph[clight];
					affectedTail = graph[clight];
				}
				break;
			}
			runner = runner->nextSibling;
		}
	}
	floodOutwards(mightBeAffected, affectedTail);
}

bool VVTrace::sameObject()
{
	bool sameobject = true;
	for (int c = 0; c < deepBuffer[crow][ccol].numCorners; c++) { //if numCorners == 0 then sameobject = true
		sameobject &= (deepBuffer[crow][ccol].corners[c]->objectID == deepBuffer[crow][ccol].objectID);
	}
	if (!sameobject && !flags[USE_GRID]) {
		return false;
	}
	if (!sameobject && flags[USE_3OBJECT]) {
		//try to find 3 same
		if (deepBuffer[crow][ccol].numCorners < 4) {
			sameobject = false;
		}
		int centerObject = deepBuffer[crow][ccol].objectID;
		for (int c = 0; c < deepBuffer[crow][ccol].numCorners; c++) {
			if (deepBuffer[crow][ccol].corners[c]->objectID != centerObject) {
				if (deepBuffer[crow][ccol].numCorners == 3) {
					sameobject = false;
				} else {
					deepBuffer[crow][ccol].corners[c] = deepBuffer[crow][ccol].corners[3];
					deepBuffer[crow][ccol].numCorners = 3;
					c--;
				}
			}
		}
	}
	return sameobject;
}


bool VVTrace::highCurvature(int row, int col) {
	Point3d left;
	Point3d center; 
	Point3d right;
	int c1, c2;
	int maxdist = 0;
	int distr, distc;
	if (deepBuffer[row][col].numCorners < 3) {
		return false;
	}
	P3D_Assign_Vector(center, deepBuffer[row][col].normal);

	for (int c = 0; c < deepBuffer[row][col].numCorners-1; c++) {
		for (int c_ = 1; c_ < deepBuffer[row][col].numCorners; c_++) {
			distr = deepBuffer[row][col].corners[c]->row - deepBuffer[row][col].corners[c_]->row;
			distr = distr*distr;
			distc = deepBuffer[row][col].corners[c]->col - deepBuffer[row][col].corners[c_]->col;
			distc = distc*distc;
			if (distc + distr > maxdist) {
				c1 = c;
				c2 = c_;
				maxdist = distc+distr;
			}
		}
	}
	P3D_Assign_Vector(left, deepBuffer[row][col].corners[c1]->normal);
	P3D_Assign_Vector(right, deepBuffer[row][col].corners[c2]->normal);


	float dotSides, dotLeft, dotRight;

	P3D_VDot(dotSides, left, right);
	P3D_VDot(dotLeft, left, center);
	P3D_VDot(dotRight, right, center);
	if (dotRight > 0.99999 && dotLeft > 0.99999) {
		return false;
	}
	if ((dotSides+0.000001 > dotLeft) || (dotSides+0.000001 > dotRight)) {
		return true;
	} else {
		return false;
	}
}

void VVTrace::writePPM(char *filename)
{
	FILE * fp = fopen(filename, "wb");
	if (fp != NULL) {
		fprintf(fp, "P6 %d %d %d", size, size, 255);
		int val;
		for (int r = 0; r < size; r++) {
			for (int c = 0; c < size; c++) {
				val = (int) (255*traced[r][c][2]);
				if (val > 255) {
					val = 255;
				}
				putc(val, fp);
				val = (int) (255*traced[r][c][0]);
				if (val > 255) {
					val = 255;
				}
				putc(val, fp);
				val = (int) (255*traced[r][c][1]);
				if (val > 255) {
					val = 255;
				}
				putc(val, fp);
			}
		}
		fclose(fp);
	}
}

void VVTrace::writeDiffPPM(char *filename, VVTrace* other)
{
	FILE * fp = fopen(filename, "wb");
	if (fp != NULL) {
		fprintf(fp, "P6 %d %d %d", size, size, 255);
		int val, sum;
		for (int r = 0; r < size; r++) {
			for (int c = 0; c < size; c++) {
				sum = 0;
				val = abs((int)(255*(traced[r][c][2] - other->traced[r][c][2])));
				if (val > 255) {
					val = 255;
				}
				sum += val;
				putc(val, fp);
				val = abs((int)(255*(traced[r][c][0] - other->traced[r][c][0])));
				if (val > 255) {
					val = 255;
				}
				sum += val;
				putc(val, fp);
				val = abs((int)(255*(traced[r][c][1] - other->traced[r][c][1])));
				if (val > 255) {
					val = 255;
				}
				sum += val;
				if (sum == 0) {
					putc(255, fp);
				} else {
					putc(val, fp);
				}
			}
		}
		fclose(fp);
	}
}

//uses the blocker at clight from ccorner to add to certainties of all lights at crow,ccol
void VVTrace::addSingleBlockage(Blocker_List* possibleOccluder) {
	DeepPixel& dp = deepBuffer[crow][ccol];
	Point3d sum;
	float bestdot;
	int centerLightNum; //the light that this occluder falls into
	float dist;

	P3D_Assign_Vector(sum, possibleOccluder->occluder);
	P3D_VSubEq(sum, dp.intersection);
	P3D_VNormalizeEq(sum);
	centerLightNum = Container::findNearestLight(sum, true);

	P3D_VDot(dist, sum, thelights[centerLightNum].direction);

	if (sum[1] > 0) {
		cblocked[centerLightNum]++;
	}

	bestdot = dist;
	if (bestdot > bestfall[centerLightNum]) {					
		bestfall[centerLightNum] = bestdot;
		bestoccluder[centerLightNum] = possibleOccluder;
	}
}

void VVTrace::addOccluder(int row, int col, int light, Point3d &occluder)
{
	Blocker_List* temp = new Blocker_List();
	P3D_Assign_Vector(temp->occluder, occluder);
	temp->light = light;
	temp->next = deepBuffer[row][col].occluders;
	deepBuffer[row][col].occluders = temp;
}

void VVTrace::removeOccluder(int row, int col, int light) {
	Blocker_List* runner = NULL;
	Blocker_List* behind = deepBuffer[row][col].occluders;
	if (behind != NULL && behind->light == light) {
		deepBuffer[row][col].occluders = behind->next;
		delete behind;
		return;
	}
	while (behind != NULL) {
		runner = behind->next;
		if (runner != NULL && runner->light == light) {
			behind->next = runner->next;
			delete runner;
			return;
		}
		behind = behind->next;
	}
}

bool VVTrace::notEnoughCorners(int numcorners)
{	
	return ((flags[USE_GRID] && flags[USE_3OBJECT] && numcorners < 3) ||
	(flags[USE_GRID] && !flags[USE_3OBJECT] && numcorners < 4) ||
	(!flags[USE_GRID] && !flags[USE_ONEBACK] && numcorners < 3) ||
	(!flags[USE_GRID] && flags[USE_ONEBACK] && numcorners < 1) ||
	(flags[USE_TRUE]));

}

void VVTrace::floodHorizon()
{
	Light_Graph* mightBeAffected = NULL;
	Light_Graph* affectedTail = NULL;
	Light_Graph* runner = NULL;
	for (int h = 0; h < numlights; h++) {
		graph[h]->nextSibling = NULL;
	}
	for (clight = 0; clight < numlights; clight++) { //add horizon lights
		if (deepBuffer[crow][ccol].prediction[clight] == PREDVIS ||
			deepBuffer[crow][ccol].prediction[clight] == PREDINVIS) {
			if (Container::horizons[clight]) {
				if (mightBeAffected == NULL) {
					mightBeAffected = graph[clight];
					affectedTail = mightBeAffected;
				}
				if (graph[clight]->nextSibling == NULL && graph[clight] != affectedTail) {
					affectedTail->nextSibling = graph[clight];
					affectedTail = graph[clight];
				}
			} else {
				runner = graph[clight]->firstChild;
				while (runner != NULL) {
					if (deepBuffer[crow][ccol].prediction[runner->lightNumber] == BHORIZON) {
						if (mightBeAffected == NULL) {
							mightBeAffected = graph[clight];
							affectedTail = mightBeAffected;
						}
						if (graph[clight]->nextSibling == NULL && graph[clight] != affectedTail) {
							affectedTail->nextSibling = graph[clight];
							affectedTail = graph[clight];
						}
						break;
					}
					runner = runner->nextSibling;
				}
			}
		}
	}
	floodOutwards(mightBeAffected, affectedTail);
}

void VVTrace::retraceTraced()
{
	Point3d occluder;
	for (crow = 0; crow < size; crow++) {
		for (ccol = 0; ccol < size; ccol++) {
			DeepPixel& dp = deepBuffer[crow][ccol];
			if (dp.prediction == NULL) {
				continue;
			}
			for (clight = 0; clight < numlights; clight++) {
			if (dp.prediction[clight] == WASTRACEDVIS ||
				dp.prediction [clight] == WASTRACEDINVIS) {
					traceShadowRay(occluder);
				}
			}
		}
	}
}

void VVTrace::traceStrongestSources()
{
	int source;
	int s = 0;
	Light_Graph* runner;
	char alsopred;
	while (Container::strongestSources[s] != -1) {
		source = Container::strongestSources[s];
		char& pred = deepBuffer[crow][ccol].prediction[source];
		if (pred == PREDVIS) {
			runner = graph[source]->firstChild;
			while (runner != NULL) {
				alsopred = deepBuffer[crow][ccol].prediction[runner->lightNumber];
				if (alsopred == WASTRACEDINVIS || alsopred == PREDINVIS) {
					pred = TOBETRACED;
					break;
				}
				runner = runner->nextSibling;
			}
		}
		s++;
	}
}

void VVTrace::fastRender()
{
	int row, col, light, object;
	float ndotl;
	for (row = 0; row < size; row++) {
		for (col = 0; col < size; col++) {
			object = Image_Info[row][col].objectID;
			if (object == 0) continue;
			Point3d& normal = Image_Info[row][col].iNormal;
			Point3d& tracedpixel = traced[row][col];
			tracedpixel[0] = 0;
			tracedpixel[1] = 0;
			tracedpixel[2] = 0;
			for (light = 0; light < numlights; light++) {
				if (Image_Info[row][col].depths[light] > 0) continue;
				ndotl = thelights[light].direction[0] * normal[0] +
						thelights[light].direction[1] * normal[0] +
						thelights[light].direction[2] * normal[0];
				if (ndotl < 0) continue;
				tracedpixel[0] += thelights[light].color[0] * ndotl * Container::objectColor[object][0];
				tracedpixel[1] += thelights[light].color[1] * ndotl * Container::objectColor[object][1];
				tracedpixel[2] += thelights[light].color[2] * ndotl * Container::objectColor[object][2];
			}
		}
	}
}
